import { EditorView } from '@codemirror/view';
import { Text, ChangeDesc, EditorStateConfig, EditorState } from '@codemirror/state';

/**
A changed range.
*/
declare class Change {
    /**
    The start of the change in document A.
    */
    readonly fromA: number;
    /**
    The end of the change in document A. This is equal to `fromA`
    in case of insertions.
    */
    readonly toA: number;
    /**
    The start of the change in document B.
    */
    readonly fromB: number;
    /**
    The end of the change in document B. This is equal to `fromB`
    for deletions.
    */
    readonly toB: number;
    constructor(
    /**
    The start of the change in document A.
    */
    fromA: number, 
    /**
    The end of the change in document A. This is equal to `fromA`
    in case of insertions.
    */
    toA: number, 
    /**
    The start of the change in document B.
    */
    fromB: number, 
    /**
    The end of the change in document B. This is equal to `fromB`
    for deletions.
    */
    toB: number);
}
/**
Compute the difference between two strings.
*/
declare function diff(a: string, b: string): readonly Change[];
/**
Compute the difference between the given strings, and clean up the
resulting diff for presentation to users by dropping short
unchanged ranges, and aligning changes to word boundaries when
appropriate.
*/
declare function presentableDiff(a: string, b: string): readonly Change[];

/**
A chunk describes a range of lines which have changed content in
them. Either side (a/b) may either be empty (when its `to` is
equal to its `from`), or points at a range starting at the start
of the first changed line, to 1 past the end of the last changed
line. Note that `to` positions may point past the end of the
document. Use `endA`/`endB` if you need an end position that is
certain to be a valid document position.
*/
declare class Chunk {
    /**
    The individual changes inside this chunk. These are stored
    relative to the start of the chunk, so you have to add
    `chunk.fromA`/`fromB` to get document positions.
    */
    readonly changes: readonly Change[];
    /**
    The start of the chunk in document A.
    */
    readonly fromA: number;
    /**
    The end of the chunk in document A. This is equal to `fromA`
    when the chunk covers no lines in document A, or is one unit
    past the end of the last line in the chunk if it does.
    */
    readonly toA: number;
    /**
    The start of the chunk in document B.
    */
    readonly fromB: number;
    /**
    The end of the chunk in document A.
    */
    readonly toB: number;
    constructor(
    /**
    The individual changes inside this chunk. These are stored
    relative to the start of the chunk, so you have to add
    `chunk.fromA`/`fromB` to get document positions.
    */
    changes: readonly Change[], 
    /**
    The start of the chunk in document A.
    */
    fromA: number, 
    /**
    The end of the chunk in document A. This is equal to `fromA`
    when the chunk covers no lines in document A, or is one unit
    past the end of the last line in the chunk if it does.
    */
    toA: number, 
    /**
    The start of the chunk in document B.
    */
    fromB: number, 
    /**
    The end of the chunk in document A.
    */
    toB: number);
    /**
    Returns `fromA` if the chunk is empty in A, or the end of the
    last line in the chunk otherwise.
    */
    get endA(): number;
    /**
    Returns `fromB` if the chunk is empty in B, or the end of the
    last line in the chunk otherwise.
    */
    get endB(): number;
    /**
    Build a set of changed chunks for the given documents.
    */
    static build(a: Text, b: Text): readonly Chunk[];
    /**
    Update a set of chunks for changes in document A. `a` should
    hold the updated document A.
    */
    static updateA(chunks: readonly Chunk[], a: Text, b: Text, changes: ChangeDesc): readonly Chunk[];
    /**
    Update a set of chunks for changes in document B.
    */
    static updateB(chunks: readonly Chunk[], a: Text, b: Text, changes: ChangeDesc): readonly Chunk[];
}

/**
Configuration options to `MergeView` that can be provided both
initially and to [`reconfigure`](https://codemirror.net/6/docs/ref/#merge.MergeView.reconfigure).
*/
interface MergeConfig {
    /**
    Controls whether editor A or editor B is shown first. Defaults
    to `"a-b"`.
    */
    orientation?: "a-b" | "b-a";
    /**
    Controls whether revert controls are shown between changed
    chunks.
    */
    revertControls?: "a-to-b" | "b-to-a";
    /**
    When given, this function is called to render the button to
    revert a chunk.
    */
    renderRevertControl?: () => HTMLElement;
    /**
    By default, the merge view will mark inserted and deleted text
    in changed chunks. Set this to false to turn that off.
    */
    highlightChanges?: boolean;
    /**
    Controls whether a gutter marker is shown next to changed lines.
    */
    gutter?: boolean;
    /**
    When given, long stretches of unchanged text are collapsed.
    `margin` gives the number of lines to leave visible after/before
    a change (default is 3), and `minSize` gives the minimum amount
    of collapsible lines that need to be present (defaults to 4).
    */
    collapseUnchanged?: {
        margin?: number;
        minSize?: number;
    };
}
/**
Configuration options given to the [`MergeView`](https://codemirror.net/6/docs/ref/#merge.MergeView)
constructor.
*/
interface DirectMergeConfig extends MergeConfig {
    /**
    Configuration for the first editor (the left one in a
    left-to-right context).
    */
    a: EditorStateConfig;
    /**
    Configuration for the second editor.
    */
    b: EditorStateConfig;
    /**
    Parent element to append the view to.
    */
    parent?: Element | DocumentFragment;
    /**
    An optional root. Only necessary if the view is mounted in a
    shadow root or a document other than the global `document`
    object.
    */
    root?: Document | ShadowRoot;
}
/**
A merge view manages two editors side-by-side, highlighting the
difference between them and vertically aligning unchanged lines.
If you want one of the editors to be read-only, you have to
configure that in its extensions.

By default, views are not scrollable. Style them (`.cm-mergeView`)
with a height and `overflow: auto` to make them scrollable.
*/
declare class MergeView {
    /**
    The first editor.
    */
    a: EditorView;
    /**
    The second editor.
    */
    b: EditorView;
    /**
    The outer DOM element holding the view.
    */
    dom: HTMLElement;
    private editorDOM;
    private revertDOM;
    private revertToA;
    private revertToLeft;
    private renderRevert;
    /**
    The current set of changed chunks.
    */
    chunks: readonly Chunk[];
    private measuring;
    /**
    Create a new merge view.
    */
    constructor(config: DirectMergeConfig);
    private dispatch;
    /**
    Reconfigure an existing merge view.
    */
    reconfigure(config: MergeConfig): void;
    private setupRevertControls;
    private scheduleMeasure;
    private measure;
    private updateRevertButtons;
    private renderRevertButton;
    private revertClicked;
    /**
    Destroy this merge view.
    */
    destroy(): void;
}
/**
Get the changed chunks for the merge view that this editor is part
of, plus the side it is on. Or null if the editor isn't part of a
merge view or the merge view hasn't finished initializing yet.
*/
declare function getChunks(state: EditorState): {
    chunks: readonly Chunk[];
    side: "a" | "b";
} | null;

export { Change, Chunk, DirectMergeConfig, MergeConfig, MergeView, diff, getChunks, presentableDiff };
